from ..loaders import parse_credential_map
from nornir.core import Nornir
from nornir.core.inventory import ConnectionOptions


def resolve_credentials(cred_map, cfg):
    """
    Resolving all credentials in the cred map based on their
    mapper functions
    """

    # first we need to resolve all the credentials in the map
    _resolve_cred(cred_map["defaults"], cfg)
    for custom_cred in cred_map["custom"]:
        _resolve_cred(custom_cred, cfg)


def _resolve_cred(cred, cfg):
    """
    Resolves enable and password fields in the credential map using the specified mapper
    """
    pw_keys = ["password", "enable"]
    for key in pw_keys:
        if key in cred:
            cred[key] = cred["mapper"](cred[key], cfg)

def update_nornir_credentials(nr: Nornir, cfg: dict):
    """
    Uses the credential map to update host credentials in this nornir instance
    """

    # parse and resolve credentials
    cred_map = parse_credential_map(cfg)
    resolve_credentials(cred_map, cfg)

    # set nornir default username and password
    nr.inventory.defaults.username = cred_map["defaults"]["username"]
    nr.inventory.defaults.password = cred_map["defaults"]["password"]
    if cred_map["defaults"]["enable"]:
        conn_options = ConnectionOptions(
            extras={"optional_args": {"secret": cred_map["defaults"]["enable"]}}
        )
        nr.inventory.defaults.connection_options["umnet_napalm"] = conn_options

    # for each host in the inventory, if there's a custom
    # match, set the credentials for that host
    for host in nr.inventory.hosts.values():

        # we only want to match a custom credential once for a particular host
        # the first match on the list
        found_custom_cred = False

        for custom_cred in cred_map["custom"]:
            if custom_cred["inventory_filter"](host) and not found_custom_cred:

                found_custom_cred = True
                host.username = custom_cred["username"]
                host.password = custom_cred["password"]

                if custom_cred.get("enable"):
                    conn_options = ConnectionOptions(
                        extras={"optional_args": {"secret": custom_cred["enable"]}}
                    )
                    host.connection_options["umnet_napalm"] = conn_options



from typing import Optional, Dict
import ipaddress
import importlib.machinery
import importlib.util

from os import getenv
import re

from git import Repo
from decouple import UndefinedValueError, Config, RepositoryEnv

from nornir.core.inventory import Host


class AgadorError(Exception):
    pass


def is_ip_address(ip: str) -> bool:
    """
    Returns whether a string is an IP (eg 10.233.0.10 or fe80::1)
    """
    try:
        ipaddress.ip_address(ip)
    except ValueError:
        return False
    return True


def is_ip_or_prefix(ip: str) -> bool:
    """
    Returns whether a string is an ip (10.233.0.10) or
    IP + prefix (10.233.0.10/24). IPv6 works as well.
    """
    try:
        ipaddress.ip_interface(ip)
    except ValueError:
        return False
    return True


def validate_email_list(input_str: str):
    """
    Makes sure that the input string is a comma-separated list
    of email addresses. Fun fact - the 'local' part of the email address
    can have a super wide range of characters, so we're just checking for non-whitespace
    The domain can have a-zA-Z0-9 and -, so we can be a bit more circumspect there.

    Function does nothing but throw an exception if the email is invalid
    """
    for email in re.split(r",", input_str):
        if not re.match(r"\S+@[\w\-.]+\.\w+$", email):
            raise ValueError(f"Invalid email {email}")


def git_update(repo_path: str, commit_message: str):
    """
    Updates git repo and if applicable, the remote origin.
    Will overwrite origin.
    """
    repo = Repo(repo_path)
    repo.git.status()

    repo.git.add(all=True)
    repo.index.commit(commit_message)
    try:
        origin = repo.remote(name="origin")
    except ValueError:
        origin = None

    if origin:
        origin.push()


def get_device_cmd_list(
    cmd_map: dict, host: Host, cmd_list_filter: Optional[list] = None
) -> Dict[str, str]:
    """
    Gets list of commands tied to a device based on its
    host inventory data. Optionally provide a list of commands to
    restrict the output to.

    Returns a dict mapping the agador command to the umnet_napalm getter
    """
    cmd_list = {}
    for cmd, data in cmd_map.items():

        if cmd_list_filter and cmd not in cmd_list_filter:
            continue

        if not data.get("inventory_filter") or data["inventory_filter"](host):
            cmd_list[cmd] = data["getter"]

    return cmd_list
